@@ -2,7 +2,7 @@
 from .coverage import CoverageTracer
 from .future_tracer import FutureTracer
 from .lazy_imports import LazyImportTracer
-from .null_coalesce import NullCoalescer
+from .optional_chaining import OptionalChainer
 from .quasiquote import Quasiquoter
 from .quick_lambda import QuickLambdaTracer
 
@@ -11,7 +11,7 @@ __all__ = [
     "CoverageTracer",
     "FutureTracer",
     "LazyImportTracer",
-    "NullCoalescer",
+    "OptionalChainer",
     "Quasiquoter",
     "QuickLambdaTracer",
 ]
