@@ -3,7 +3,16 @@ import ast
 import logging
 import sys
 from contextlib import contextmanager
-from typing import cast, Dict, FrozenSet, List, Optional, Set, Union
+from typing import (
+    cast,
+    TYPE_CHECKING,
+    Callable,
+    DefaultDict,
+    Dict,
+    List,
+    Optional,
+    Union,
+)
 
 from pyccolo import fast
 from pyccolo.extra_builtins import TRACING_ENABLED, make_guard_name
@@ -15,6 +24,10 @@ from pyccolo.fast import (
 )
 from pyccolo.trace_events import TraceEvent
 
+if TYPE_CHECKING:
+    from pyccolo.ast_rewriter import GUARD_DATA_T
+    from pyccolo.tracer import BaseTracer
+
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.WARNING)
@@ -23,29 +36,36 @@ logger.setLevel(logging.WARNING)
 class ExprRewriter(ast.NodeTransformer, EmitterMixin):
     def __init__(
         self,
+        tracers: "List[BaseTracer]",
         orig_to_copy_mapping: Dict[int, ast.AST],
-        events_with_handlers: FrozenSet[TraceEvent],
-        guards: Set[str],
+        handler_predicate_by_event: DefaultDict[TraceEvent, Callable[..., bool]],
+        handler_guards_by_event: DefaultDict[TraceEvent, List["GUARD_DATA_T"]],
     ):
-        EmitterMixin.__init__(self, orig_to_copy_mapping, events_with_handlers, guards)
+        EmitterMixin.__init__(
+            self,
+            tracers,
+            orig_to_copy_mapping,
+            handler_predicate_by_event,
+            handler_guards_by_event,
+        )
         self._top_level_node_for_symbol: Optional[ast.AST] = None
 
     def visit(self, node: ast.AST):
         ret = super().visit(node)
         if isinstance(node, ast.stmt):
-            # we haven't inserted statements yet, and StatementInserter needs the previous ids to be identical
+            # We haven't inserted statements yet, and StatementInserter
+            # needs the previous ids to be identical.
             assert ret is node
         return ret
 
     def visit_Name(self, node: ast.Name):
-        if (
-            isinstance(node.ctx, ast.Load)
-            and TraceEvent.load_name in self.events_with_handlers
-        ):
-            with fast.location_of(node):
-                return self.emit(TraceEvent.load_name, node, ret=node)
-        else:
+        if not isinstance(node.ctx, ast.Load):
             return node
+        orig_node = node
+        if self.handler_predicate_by_event[TraceEvent.load_name](orig_node):
+            with fast.location_of(orig_node):
+                node = self.emit(TraceEvent.load_name, orig_node, ret=node)  # type: ignore
+        return node
 
     @contextmanager
     def attrsub_context(self, top_level_node: Optional[ast.AST]):
@@ -115,6 +135,12 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
                 keywords=[],
             )
         elif isinstance(subscript, (ast.ExtSlice, ast.Tuple)):
+            if isinstance(subscript, ast.Tuple):
+                elts = subscript.elts
+            else:
+                elts = subscript.dims  # type: ignore
+            elts = [self._maybe_convert_ast_subscript(elt) for elt in elts]
+            subscript = fast.Tuple(elts, ast.Load())
             return cast(ast.expr, subscript)
         else:
             return self.visit(subscript)
@@ -126,13 +152,15 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
                 node.slice if hasattr(node.slice, "lineno") else node.value
             ):
                 slc = self._maybe_convert_ast_subscript(node.slice)
-                if isinstance(slc, (ast.ExtSlice, ast.Tuple)):
-                    elts = slc.elts if isinstance(slc, ast.Tuple) else slc.dims  # type: ignore
-                    elts = [self._maybe_convert_ast_subscript(elt) for elt in elts]  # type: ignore
-                    slc = fast.Tuple(elts, ast.Load())
-                if TraceEvent.subscript_slice in self.events_with_handlers:
-                    slc = self.emit(TraceEvent.subscript_slice, node, ret=slc)
-                if evt_to_use in self.events_with_handlers:
+                if self.handler_predicate_by_event[TraceEvent.before_subscript_slice](
+                    node
+                ):
+                    slc = self.emit(TraceEvent.before_subscript_slice, node, ret=slc)
+                if self.handler_predicate_by_event[TraceEvent.after_subscript_slice](
+                    node
+                ):
+                    slc = self.emit(TraceEvent.after_subscript_slice, node, ret=slc)
+                if self.handler_predicate_by_event[evt_to_use](node.slice):
                     replacement_slice: ast.expr = self.emit(
                         TraceEvent._load_saved_slice, node.slice
                     )
@@ -151,14 +179,11 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
         node,
         call_context=False,
         orig_node_id=None,
-        begin_kwargs=None,
-        end_kwargs=None,
     ):
         if self._inside_attrsub_load_chain:
             return node
         orig_node = node
         orig_node_id = orig_node_id or id(orig_node)
-        end_kwargs = end_kwargs or {}
 
         ctx = getattr(orig_node, "ctx", ast.Load())
         is_load = isinstance(ctx, ast.Load)
@@ -166,18 +191,24 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
             return node
 
         with fast.location_of(node):
-            end_kwargs["call_context"] = fast.NameConstant(call_context)
-            if TraceEvent.before_load_complex_symbol in self.events_with_handlers:
-                node = self.make_tuple_event_for(
-                    node,
-                    TraceEvent.before_load_complex_symbol,
-                    orig_node_id=orig_node_id,
-                    **(begin_kwargs or {}),
-                )
-            if TraceEvent.after_load_complex_symbol in self.events_with_handlers:
-                end_kwargs["ret"] = node
+            extra_kwargs = dict(call_context=fast.NameConstant(call_context))
+            if self.handler_predicate_by_event[TraceEvent.before_load_complex_symbol](
+                orig_node
+            ):
                 node = self.emit(
-                    TraceEvent.after_load_complex_symbol, orig_node_id, **end_kwargs
+                    TraceEvent.before_load_complex_symbol,
+                    orig_node_id,
+                    ret=node,
+                    **extra_kwargs,
+                )
+            if self.handler_predicate_by_event[TraceEvent.after_load_complex_symbol](
+                orig_node
+            ):
+                node = self.emit(
+                    TraceEvent.after_load_complex_symbol,
+                    orig_node_id,
+                    ret=node,
+                    **extra_kwargs,
                 )
         # end location_of(node)
         return node
@@ -189,10 +220,11 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
         evt_to_use: TraceEvent,
         call_context: bool = False,
     ):
-        orig_node_id = id(node)
+        orig_node = node
+        orig_node_id = id(orig_node)
         with fast.location_of(node.value):
             is_subscript = isinstance(node, ast.Subscript)
-            should_emit_evt = evt_to_use in self.events_with_handlers
+            should_emit_evt = self.handler_predicate_by_event[evt_to_use](orig_node)
             with self.attrsub_context(node):
                 extra_keywords: Dict[str, ast.AST] = {}
                 if isinstance(node.value, ast.Name):
@@ -204,7 +236,8 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
                         if isinstance(node, ast.Subscript):
                             slice_val = subscript_to_slice(node)
                             if isinstance(slice_val, ast.Name):
-                                # TODO: this should be more general than just simple ast.Name subscripts
+                                # TODO: this should be more general than
+                                #  just simple ast.Name subscripts
                                 subscript_name = slice_val.id
                         extra_keywords["subscript_name"] = (
                             fast.NameConstant(None)
@@ -229,9 +262,9 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
                 if is_subscript
                 else TraceEvent.after_attribute_load
             )
-            if after_evt in self.events_with_handlers:
+            if self.handler_predicate_by_event[after_evt](orig_node):
                 with fast.location_of(node):
-                    node = self.emit(
+                    node = self.emit(  # type: ignore
                         after_evt,
                         orig_node_id,
                         ret=node,
@@ -254,7 +287,7 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
             with fast.location_of(maybe_kwarg):
                 with self.attrsub_context(None):
                     new_arg_value = self.visit(maybe_kwarg)
-                if TraceEvent.argument in self.events_with_handlers:
+                if self.handler_predicate_by_event[TraceEvent.argument](maybe_kwarg):
                     with self.attrsub_context(None):
                         new_arg_value = cast(
                             ast.expr,
@@ -274,21 +307,27 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
         return replacement_args
 
     def visit_Call(self, node: ast.Call):
+        ret_as_call: ast.Call = node
+        ret: Union[ast.Call, ast.IfExp] = node
         orig_node_id = id(node)
 
-        with self.attrsub_context(node):
-            if isinstance(node.func, ast.Attribute):
-                node.func = self.visit_Attribute(node.func, call_context=True)
-            elif isinstance(node.func, ast.Subscript):
-                node.func = self.visit_Subscript(node.func, call_context=True)
+        with self.attrsub_context(ret_as_call):
+            if isinstance(ret_as_call.func, ast.Attribute):
+                ret_as_call.func = self.visit_Attribute(
+                    ret_as_call.func, call_context=True
+                )
+            elif isinstance(ret_as_call.func, ast.Subscript):
+                ret_as_call.func = self.visit_Subscript(
+                    ret_as_call.func, call_context=True
+                )
             else:
-                node.func = self.visit(node.func)
+                ret_as_call.func = self.visit(ret_as_call.func)
 
         # TODO: need a way to rewrite ast of subscript args,
         #  and to process these separately from outer rewrite
 
-        node.args = self._get_replacement_args(node.args, False)
-        node.keywords = self._get_replacement_args(node.keywords, True)
+        ret_as_call.args = self._get_replacement_args(ret_as_call.args, False)
+        ret_as_call.keywords = self._get_replacement_args(ret_as_call.keywords, True)
 
         # in order to ensure that the args are processed with appropriate active scope,
         # we need to make sure not to use the active namespace scope on args (in the case
@@ -300,47 +339,82 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
         #
         # This effectively rewrites function calls as follows:
         # f(a, b, ..., c) -> trace(f, 'enter argument list')(a, b, ..., c)
-        if TraceEvent.before_call in self.events_with_handlers:
-            with fast.location_of(node.func):
-                node.func = self.emit(
+        if self.handler_predicate_by_event[TraceEvent.before_call](node):
+            with fast.location_of(ret_as_call.func):
+                ret_as_call.func = self.emit(
                     TraceEvent.before_call,
                     orig_node_id,
-                    ret=node.func,
+                    ret=ret_as_call.func,
                     call_node_id=self.get_copy_id_ast(orig_node_id),
                 )
 
         # f(a, b, ..., c) -> trace(f(a, b, ..., c), 'exit argument list')
-        if TraceEvent.after_call in self.events_with_handlers:
-            with fast.location_of(node):
-                node = self.emit(
+        if self.handler_predicate_by_event[TraceEvent.after_call](node):
+            with fast.location_of(ret):
+                ret = self.emit(
                     TraceEvent.after_call,
                     orig_node_id,
-                    ret=node,
+                    ret=ret,
                     call_node_id=self.get_copy_id_ast(orig_node_id),
                 )
+                del ret_as_call
 
         return self._maybe_wrap_symbol_in_before_after_tracing(
-            node, call_context=True, orig_node_id=orig_node_id
+            ret, call_context=True, orig_node_id=orig_node_id
         )
 
+    def visit_Expr(self, node: ast.Expr) -> ast.Expr:
+        with fast.location_of(node):
+            node.value = self.visit(node.value)
+            if self.handler_predicate_by_event[TraceEvent.after_expr_stmt](node):
+                node.value = self.emit(TraceEvent.after_expr_stmt, node, ret=node.value)
+            return node
+
+    def visit_With(self, node: ast.With) -> ast.With:
+        if self.is_tracing_disabled_context(node):
+            return node
+        else:
+            return cast(ast.With, self.generic_visit(node))
+
     def visit_Assign(self, node: ast.Assign):
-        new_targets = []
-        for target in node.targets:
-            new_targets.append(self.visit(target))
-        node.targets = new_targets
-        orig_value_id = id(node.value)
+        return self.visit_Assign_or_AugAssign(node)
+
+    def visit_AugAssign(self, node: ast.AugAssign):
+        return self.visit_Assign_or_AugAssign(node)
+
+    def visit_Assign_or_AugAssign(self, node: Union[ast.Assign, ast.AugAssign]):
+        if isinstance(node, ast.Assign):
+            new_targets = []
+            for target in node.targets:
+                new_targets.append(self.visit(target))
+            node.targets = new_targets
+        else:
+            node.target = self.visit(node.target)
+        orig_value = node.value
+        orig_value_id = id(orig_value)
+        if isinstance(node, ast.Assign):
+            before_evt = TraceEvent.before_assign_rhs
+            after_evt = TraceEvent.after_assign_rhs
+        else:
+            before_evt = TraceEvent.before_augassign_rhs
+            after_evt = TraceEvent.after_augassign_rhs
         with fast.location_of(node.value):
             node.value = self.visit(node.value)
-            if TraceEvent.before_assign_rhs in self.events_with_handlers:
-                node.value = self.make_tuple_event_for(
-                    node.value, TraceEvent.before_assign_rhs, orig_node_id=orig_value_id
-                )
-            if TraceEvent.after_assign_rhs in self.events_with_handlers:
-                node.value = self.emit(
-                    TraceEvent.after_assign_rhs, orig_value_id, ret=node.value
-                )
+            if self.handler_predicate_by_event[before_evt](orig_value):
+                node.value = self.emit(before_evt, orig_value_id, ret=node.value)
+            if self.handler_predicate_by_event[after_evt](orig_value):
+                node.value = self.emit(after_evt, orig_value_id, ret=node.value)
         return node
 
+    def visit_If(self, node: ast.If):
+        with fast.location_of(node):
+            node.test = self.visit(node.test)
+            node.body = [self.visit(stmt) for stmt in node.body]
+            node.orelse = [self.visit(stmt) for stmt in node.orelse]
+            if self.handler_predicate_by_event[TraceEvent.after_if_test](node):
+                node.test = self.emit(TraceEvent.after_if_test, node, ret=node.test)
+            return node
+
     def visit_Lambda(self, node: ast.Lambda):
         assert isinstance(getattr(node, "ctx", ast.Load()), ast.Load)
         untraced_lam = cast(ast.Lambda, self.orig_to_copy_mapping[id(node)])
@@ -355,38 +429,54 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
                             node,
                             ret=fast.NameConstant(True),
                         )
-                        if TraceEvent.before_lambda_body in self.events_with_handlers
+                        if self.handler_predicate_by_event[
+                            TraceEvent.before_lambda_body
+                        ](untraced_lam)
                         else None,
                     ]
                 ),
                 body=ret_node.body,
                 orelse=untraced_lam.body,
             )
-            if TraceEvent.before_lambda in self.events_with_handlers:
-                ret_node = self.make_tuple_event_for(
-                    ret_node, TraceEvent.before_lambda, orig_node_id=id(node)
+            if self.handler_predicate_by_event[TraceEvent.after_lambda_body](
+                untraced_lam
+            ):
+                ret_node.body = self.emit(
+                    TraceEvent.after_lambda_body, node, ret=ret_node.body
                 )
-            if TraceEvent.after_lambda in self.events_with_handlers:
-                ret_node = self.emit(TraceEvent.after_lambda, node, ret=ret_node)
+            if self.handler_predicate_by_event[TraceEvent.before_lambda](untraced_lam):
+                ret_node = self.emit(TraceEvent.before_lambda, node, ret=ret_node)  # type: ignore
+            if self.handler_predicate_by_event[TraceEvent.after_lambda](untraced_lam):
+                ret_node = self.emit(TraceEvent.after_lambda, node, ret=ret_node)  # type: ignore
         return ret_node
 
     def visit_While(self, node: ast.While):
         for name, field in ast.iter_fields(node):
             if name == "test":
                 loop_node_copy = cast(ast.While, self.orig_to_copy_mapping[id(node)])
-                loop_guard = make_guard_name(loop_node_copy)
-                self.register_guard(loop_guard)
                 with fast.location_of(node):
-                    node.test = fast.IfExp(
-                        test=make_composite_condition(
-                            [
-                                make_test(TRACING_ENABLED),
-                                make_test(loop_guard),
-                            ]
-                        ),
-                        body=self.visit(field),
-                        orelse=loop_node_copy.test,
-                    )
+                    visited_test = self.visit(field)
+                    if self.handler_predicate_by_event[TraceEvent.after_while_test](
+                        node
+                    ):
+                        visited_test = self.emit(
+                            TraceEvent.after_while_test, node, ret=visited_test
+                        )
+                    if self.global_guards_enabled:
+                        loop_guard = make_guard_name(loop_node_copy)
+                        self.register_guard(loop_guard)
+                        node.test = fast.IfExp(
+                            test=make_composite_condition(
+                                [
+                                    make_test(TRACING_ENABLED),
+                                    make_test(loop_guard),
+                                ]
+                            ),
+                            body=visited_test,
+                            orelse=loop_node_copy.test,
+                        )
+                    else:
+                        node.test = visited_test
             elif isinstance(field, list):
                 setattr(node, name, [self.visit(elt) for elt in field])
             else:
@@ -429,6 +519,7 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
         node: Union[ast.Dict, ast.List, ast.Set, ast.Tuple],
         should_inner_visit=True,
     ):
+        untraced_lit = self.orig_to_copy_mapping[id(node)]
         ret_node: ast.expr = node
         if should_inner_visit:
             ret_node = cast(ast.expr, self.generic_visit(node))
@@ -436,12 +527,10 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
             return ret_node
         with fast.location_of(node):
             lit_before_evt = self._ast_container_to_literal_trace_evt(node, before=True)
-            if lit_before_evt in self.events_with_handlers:
-                ret_node = self.make_tuple_event_for(
-                    ret_node, lit_before_evt, orig_node_id=id(node)
-                )
+            if self.handler_predicate_by_event[lit_before_evt](untraced_lit):
+                ret_node = self.emit(lit_before_evt, node, ret=ret_node)
             lit_after_evt = self._ast_container_to_literal_trace_evt(node, before=False)
-            if lit_after_evt in self.events_with_handlers:
+            if self.handler_predicate_by_event[lit_after_evt](untraced_lit):
                 ret_node = self.emit(lit_after_evt, node, ret=ret_node)
         return ret_node
 
@@ -474,11 +563,12 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
         elt_trace_evt = self._ast_container_to_elt_trace_evt(node)
         for i, elt in enumerate(node.elts):
             if isinstance(elt, ast.Starred):
-                # TODO: trace starred elts too
                 saw_starred = True
-                traced_elts.append(elt)
+                traced_elts.append(self.visit(elt))
                 continue
-            elif not is_load or elt_trace_evt not in self.events_with_handlers:
+            elif not is_load or not self.handler_predicate_by_event[elt_trace_evt](
+                node
+            ):
                 traced_elts.append(self.visit(elt))
                 continue
             with fast.location_of(elt):
@@ -504,7 +594,7 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
             else:
                 with fast.location_of(k):
                     traced_key = self.visit(k)
-                    if TraceEvent.dict_key in self.events_with_handlers:
+                    if self.handler_predicate_by_event[TraceEvent.dict_key](k):
                         traced_key = self.emit(
                             TraceEvent.dict_key,
                             k,
@@ -515,11 +605,11 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
                     traced_keys.append(traced_key)
             with fast.location_of(v):
                 if is_dict_unpack:
-                    key_node_id_ast = fast.NameConstant(None)
+                    key_node_id_ast: ast.AST = fast.NameConstant(None)
                 else:
                     key_node_id_ast = self.get_copy_id_ast(k)
                 traced_value = self.visit(v)
-                if TraceEvent.dict_value in self.events_with_handlers:
+                if self.handler_predicate_by_event[TraceEvent.dict_value](v):
                     traced_value = self.emit(
                         TraceEvent.dict_value,
                         v,
@@ -538,13 +628,15 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
         orig_node_value = node.value
         with fast.location_of(node):
             node.value = self.visit(node.value)
-            if TraceEvent.before_return in self.events_with_handlers:
-                node.value = self.make_tuple_event_for(
-                    node.value,
-                    TraceEvent.before_return,
-                    orig_node_id=id(orig_node_value),
+            if self.handler_predicate_by_event[TraceEvent.before_return](
+                orig_node_value
+            ):
+                node.value = self.emit(
+                    TraceEvent.before_return, orig_node_value, ret=node.value
                 )
-            if TraceEvent.after_return in self.events_with_handlers:
+            if self.handler_predicate_by_event[TraceEvent.after_return](
+                orig_node_value
+            ):
                 node.value = self.emit(
                     TraceEvent.after_return, orig_node_value, ret=node.value
                 )
@@ -556,37 +648,16 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
             target.ctx = ast.Del()  # type: ignore
         return ret
 
-    def visit_BinOp(self, node: ast.BinOp):
+    def visit_BinOp(self, node: ast.BinOp) -> Union[ast.BinOp, ast.Call, ast.IfExp]:
+        untraced_node = self.orig_to_copy_mapping[id(node)]
         op = node.op
-        if isinstance(op, ast.Add):
-            evt = TraceEvent.add
-        elif isinstance(op, ast.Sub):
-            evt = TraceEvent.sub
-        elif isinstance(op, ast.Mult):
-            evt = TraceEvent.mult
-        elif isinstance(op, ast.MatMult):
-            evt = TraceEvent.mat_mult
-        elif isinstance(op, ast.Div):
-            evt = TraceEvent.div
-        elif isinstance(op, ast.FloorDiv):
-            evt = TraceEvent.floor_div
-        elif isinstance(op, ast.Pow):
-            evt = TraceEvent.power
-        elif isinstance(op, ast.BitAnd):
-            evt = TraceEvent.bit_and
-        elif isinstance(op, ast.BitOr):
-            evt = TraceEvent.bit_or
-        elif isinstance(op, ast.BitXor):
-            evt = TraceEvent.bit_xor
-        else:
-            evt = None
 
         for attr, operand_evt in [
             ("left", TraceEvent.left_binop_arg),
             ("right", TraceEvent.right_binop_arg),
         ]:
             operand_node = getattr(node, attr)
-            if operand_evt in self.events_with_handlers:
+            if self.handler_predicate_by_event[operand_evt](operand_node):
                 with fast.location_of(operand_node):
                     setattr(
                         node,
@@ -598,16 +669,82 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
             else:
                 setattr(node, attr, self.visit(operand_node))
 
-        if evt in self.events_with_handlers:
+        ret: Union[ast.BinOp, ast.Call, ast.IfExp] = node
+        if self.handler_predicate_by_event[TraceEvent.before_binop](untraced_node):
             with fast.location_of(node):
-                return self.emit(evt, node, left=node.left, right=node.right)
+                ret = self.emit(
+                    TraceEvent.before_binop,
+                    node,
+                    ret=self.make_lambda(
+                        body=fast.BinOp(
+                            op=op,
+                            left=fast.Name("x", ast.Load()),
+                            right=fast.Name("y", ast.Load()),
+                        ),
+                        args=[fast.arg("x", None), fast.arg("y", None)],
+                    ),
+                    before_expr_args=[node.left, node.right],
+                )
+        if self.handler_predicate_by_event[TraceEvent.after_binop](untraced_node):
+            with fast.location_of(node):
+                ret = self.emit(TraceEvent.after_binop, node, ret=ret)
+        return ret
+
+    def visit_Compare(
+        self, node: ast.Compare
+    ) -> Union[ast.Compare, ast.Call, ast.IfExp]:
+        # TODO: this is pretty similar to BinOp above; maybe can dedup some code
+        untraced_node = self.orig_to_copy_mapping[id(node)]
+
+        if self.handler_predicate_by_event[TraceEvent.left_compare_arg](node.left):
+            with fast.location_of(node.left):
+                node.left = self.emit(
+                    TraceEvent.left_compare_arg, node.left, ret=self.visit(node.left)
+                )
         else:
-            return node
+            node.left = self.visit(node.left)
+
+        for idx, comparator in enumerate(node.comparators):
+            if self.handler_predicate_by_event[TraceEvent.compare_arg](comparator):
+                with fast.location_of(comparator):
+                    node.comparators[idx] = self.emit(
+                        TraceEvent.compare_arg, comparator, ret=self.visit(comparator)
+                    )
+            else:
+                node.comparators[idx] = self.visit(comparator)
+
+        ret: Union[ast.Compare, ast.Call, ast.IfExp] = node
+        if self.handler_predicate_by_event[TraceEvent.before_compare](untraced_node):
+            with fast.location_of(node):
+                ret = self.emit(
+                    TraceEvent.before_compare,
+                    node,
+                    ret=self.make_lambda(
+                        body=fast.Compare(
+                            ops=node.ops,
+                            left=fast.Name("x", ast.Load()),
+                            comparators=[
+                                fast.Name(f"y_{i}", ast.Load())
+                                for i in range(len(node.comparators))
+                            ],
+                        ),
+                        args=[fast.arg("x", None)]
+                        + [
+                            fast.arg(f"y_{i}", None)
+                            for i in range(len(node.comparators))
+                        ],
+                    ),
+                    before_expr_args=[node.left] + node.comparators,
+                )
+        if self.handler_predicate_by_event[TraceEvent.after_compare](untraced_node):
+            with fast.location_of(node):
+                ret = self.emit(TraceEvent.after_compare, node, ret=ret)
+        return ret
 
     if sys.version_info < (3, 8):
 
         def visit_Ellipsis(self, node: ast.Ellipsis):
-            if TraceEvent.ellipses in self.events_with_handlers:
+            if self.handler_predicate_by_event[TraceEvent.ellipses](node):
                 with fast.location_of(node):
                     return self.emit(TraceEvent.ellipses, node, ret=node)
             else:
@@ -616,7 +753,9 @@ class ExprRewriter(ast.NodeTransformer, EmitterMixin):
     else:
 
         def visit_Constant(self, node: ast.Constant):
-            if node.value is ... and TraceEvent.ellipses in self.events_with_handlers:
+            if node.value is ... and self.handler_predicate_by_event[
+                TraceEvent.ellipses
+            ](node):
                 with fast.location_of(node):
                     return self.emit(TraceEvent.ellipses, node, ret=node)
             else:
