@@ -6,6 +6,7 @@ from contextlib import contextmanager
 from importlib.abc import MetaPathFinder
 from importlib.machinery import SourceFileLoader
 from importlib.util import spec_from_loader, decode_source
+from typing import Callable, Generator
 
 from pyccolo.trace_events import TraceEvent
 
@@ -52,12 +53,14 @@ class TraceLoader(SourceFileLoader):
             source = decode_source(source_bytes)
             still_needs_decode = False
         except SyntaxError:
-            # this allows us to handle esoteric encodings that require parsing, such as future_annotations
-            # in this case, just guess that it's utf-8 encoded
+            # This allows us to handle esoteric encodings that
+            # require parsing, such as future_annotations.
+            # In this case, just guess that it's utf-8 encoded.
 
-            # this is a bit unfortunate in that it involves multiple round-trips of decoding / encoding,
-            # but it's the only way I can think of to ensure that source transformations happen in the
-            # correct order
+            # This is a bit unfortunate in that it involves multiple
+            # round-trips of decoding / encoding, but it's the only
+            # way I can think of to ensure that source transformations
+            # happen in the correct order.
             source = str(source_bytes, encoding="utf-8")
         for augmenter in self._syntax_augmenters:
             source = augmenter(source)
@@ -75,6 +78,7 @@ class TraceLoader(SourceFileLoader):
                 optimize=_optimize,
             )
         except Exception:
+            logger.exception("exception during source to code for path %s", path)
             return super().source_to_code(data, path, _optimize=_optimize)
 
 
@@ -121,21 +125,41 @@ class TraceFinder(MetaPathFinder):
         tracers_to_use = []
         import_ = TraceEvent.import_.value
         for tracer in self.tracers:
-            if tracer._file_passes_filter_impl(import_, source_path):
+            if tracer._should_instrument_file_impl(
+                source_path
+            ) or tracer._file_passes_filter_impl(import_, source_path):
                 tracers_to_use.append(tracer)
-        for tracer in tracers_to_use:
-            tracer._emit_event(import_, None, None)
-
         if len(tracers_to_use) == 0:
             return None
+        for tracer in tracers_to_use:
+            tracer._emit_event(import_, None, None)
         spec.loader = TraceLoader(tracers_to_use, spec.loader.name, spec.loader.path)
         return spec
 
 
-@contextmanager
-def patch_meta_path(tracers):
-    try:
+def patch_meta_path_non_context(tracers) -> Callable:
+    orig_meta_path_entry = None
+
+    def cleanup_callback():
+        if orig_meta_path_entry is None:
+            del sys.meta_path[0]
+        else:
+            sys.meta_path[0] = orig_meta_path_entry
+
+    if len(sys.meta_path) > 0 and isinstance(sys.meta_path[0], TraceFinder):
+        orig_meta_path_entry = sys.meta_path[0]
+        sys.meta_path[0] = TraceFinder(tracers)
+    else:
         sys.meta_path.insert(0, TraceFinder(tracers))
+    return cleanup_callback
+
+
+@contextmanager
+def patch_meta_path(tracers) -> Generator[None, None, None]:
+    cleanup_callback = None
+    try:
+        cleanup_callback = patch_meta_path_non_context(tracers)
         yield
     finally:
-        del sys.meta_path[0]
+        if cleanup_callback is not None:
+            cleanup_callback()
